from ...context import EvaluationContext
from ...types.primitives.int import IntegerType
from ..expression import Expression
from ..literals.int import IntLiteral


class SizeofExpression(Expression):
    def __init__(self, inner_expression: Expression):
        super().__init__()
        self.inner_expression = inner_expression

    def __repr__(self):
        return f"sizeof({self.inner_expression})"

    def get_type(self, context):
        return IntegerType()

    def evaluate(self, context: EvaluationContext) -> str:
        inner_type = self.inner_expression.get_type(context)
        size = inner_type.get_size()
        return IntLiteral(size).evaluate(context)



import json

from ...context import EvaluationContext
from ...context.heap.helpers import place_const_data_on_heap
from ...types.complex.string import StringType
from ..expression import Expression


class StringLiteral(Expression):
    def __init__(self, value: str) -> None:
        super().__init__()
        self._value = value

    def __repr__(self):
        return json.dumps(self._value)

    def get_type(self, context: EvaluationContext):
        return StringType()

    def evaluate(self, context: EvaluationContext) -> str:
        return place_const_data_on_heap(context, list(ord(c) for c in self._value))



from ...whitespace.snippets import END
from ..context import EvaluationContext, IssueLevel
from ..context.heap.helpers import initialize_heap
from ..context.stack.helpers import enter_first_stack
from ..debugstatement import debug_statement
from ..expressions.functions.function_call import FunctionCall
from ..types.primitives.void import VoidType
from .statement import Statement
from .void import VoidStatement


class Program(Statement):
    def __init__(self, statement: Statement, entry_function: str = "main"):
        super().__init__()
        self._statement = statement
        self._entry_function = entry_function

    def __repr__(self):
        return f"Program {self._statement}"

    def evaluate(self, context: EvaluationContext) -> str:
        # Evaluating the main statement block first is needed,
        # since that will register all functions
        statement_code = self._statement.evaluate(context)

        entry_function = context.function_registry.get_function_definition(self._entry_function)
        if not entry_function:
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Entry function '{self._entry_function}' is not defined.",
            )
            return ""
        if len(entry_function.parameters) > 0:
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Entry function '{self._entry_function}' must not have parameters.",
            )
            return ""
        if entry_function.return_type != VoidType():
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Entry function '{self._entry_function}' must have return type 'void'.",
            )
            return ""

        res = ""

        # initialize the heap
        res += initialize_heap(context)

        # set up the first stack
        res += enter_first_stack(context)

        # imitate function call to main function
        res += debug_statement(f"Calling entry function '{self._entry_function}'")
        res += VoidStatement(FunctionCall(name=self._entry_function, parameters=[])).evaluate(
            context
        )

        res += END

        res += debug_statement("Program statement:")
        res += statement_code

        return res



from ...whitespace.printstr import print_str
from ...whitespace.snippets import END, LABEL, PUSH, RETURN
from ..context import EvaluationContext, IssueLevel
from ..definitions.function import FunctionSignature
from ..expressions.expression import Expression
from ..types.primitives.void import VoidType
from .statement import Statement


class FunctionBodyStatement(Statement):

    def __init__(self, signature: FunctionSignature, body: Statement) -> None:
        super().__init__()
        self._signature = signature
        self._body = body

    def __repr__(self):
        res = f"{self._signature.return_type} "
        res += f"{self._signature.name}("
        res += ", ".join(str(param.name) for param in self._signature.parameters)
        res += f") {self._body}"
        return res

    def evaluate(self, context: EvaluationContext) -> str:
        entry_label = context.label_registry.new_label()
        context.function_registry.register_function(self._signature, entry_label)

        # Stack frame and initial variables are set up by the caller.
        # We just need to generate the body here.

        res = LABEL(entry_label)

        with context.function_registry.function_context(self._signature):
            res += self._body.evaluate(context)

        if self._signature.return_type == VoidType():
            # According to the calling convention, void function should still return a value,
            # so we push a dummy 0 before returning.
            res += PUSH(0) + RETURN
        else:
            # Non-void functions should return a value before reaching the end.
            # If the function body reaches here, it's an error in the source code.
            # TODO: Optimize this error by only having it once in the program, and jumping to it.
            res += print_str("Error: Non-void function did not return a value.\n")
            res += END

        return res


class ReturnStatement(Statement):

    def __init__(self, value: Expression | None):
        super().__init__()
        self._value = value

    def __repr__(self):
        if self._value is None:
            return "return"
        return f"return {self._value}"

    def evaluate(self, context: EvaluationContext) -> str:
        if not context.function_registry.current_function:
            context.register_issue(
                IssueLevel.ERROR,
                self,
                "Return statement outside of a function.",
            )
            return ""
        function_definition = context.function_registry.current_function

        if function_definition.return_type == VoidType():
            if self._value is not None:
                context.register_issue(
                    IssueLevel.ERROR,
                    self,
                    "Void function should not return a value.",
                )
                return ""

            return PUSH(0) + RETURN
        else:
            if self._value is None:
                context.register_issue(
                    IssueLevel.ERROR,
                    self,
                    "Non-void function must return a value.",
                )
                return ""
            value_type = self._value.get_type(context)
            if value_type != function_definition.return_type:
                context.register_issue(
                    IssueLevel.ERROR,
                    self,
                    f"Return type mismatch: expected {function_definition.return_type}, "
                    f"got {value_type}.",
                )
                return ""

            return self._value.evaluate(context) + RETURN



from abc import ABC, abstractmethod

from ..context import EvaluationContext
from ..node import AstNode
from ..types.type import Type


class Expression(AstNode, ABC):

    @abstractmethod
    def get_type(self, context: EvaluationContext) -> Type:
        pass

    @abstractmethod
    def evaluate(self, context: EvaluationContext) -> str:
        pass



from ..node import AstNode
from ..types.type import Type


def type_error(expected_type: Type, acutal_type: Type, node: AstNode) -> str:
    _ = node  # Unused for now, but could be used for more detailed error messages
    return f"Expected type {expected_type}, but got {acutal_type}."



def indent(text: str, indent_str: str = "  ") -> str:
    """Indents all lines in the given text with the given intent string."""
    lines = text.splitlines()
    indented_lines = [f"{indent_str}{line}" for line in lines]
    return "\n".join(indented_lines)



from __future__ import annotations

from dataclasses import dataclass

from ..types.type import Type
from .variable import VariableDefinition


@dataclass
class ParameterDefinition:
    name: str
    type: Type


@dataclass
class FunctionSignature:
    name: str
    parameters: list[ParameterDefinition]
    return_type: Type
    variables: list[VariableDefinition]

    def assert_correct_variable_configuration(self) -> None:
        """
        Validates that the function variable setup is correct according to the parameters.
        Each parameter will be stored into a variable by the same name. The first N variables
        must match the N parameters.
        """
        if len(self.variables) < len(self.parameters):
            raise RuntimeError(
                f"Function '{self.name}' has fewer variables ({len(self.variables)}) "
                f"than parameters ({len(self.parameters)})."
            )
        for i in range(len(self.parameters)):
            param = self.parameters[i]
            var = self.variables[i]
            if param.name != var.name:
                raise RuntimeError(
                    f"Function '{self.name}' parameter {i} name '{param.name}' does not match "
                    f"variable name '{var.name}'."
                )
            if param.type != var.type:
                raise RuntimeError(
                    f"Function '{self.name}' parameter {i} type '{param.type}' does not match "
                    f"variable type '{var.type}'."
                )

    def get_variable(self, name: str) -> VariableOffset | None:
        offset = 0
        for variable in self.variables:
            if variable.name == name:
                return VariableOffset(definition=variable, offset=offset)
            else:
                offset += variable.type.get_size()
        return None


@dataclass
class VariableOffset:
    definition: VariableDefinition
    offset: int



from ....whitespace.snippets import CALL, FETCH, PUSH, STORE, SWAP
from ...context import EvaluationContext, IssueLevel
from ...context.evaluation_context import increment_stack, reset_stack
from ...context.heap.consts import FUNCTION_PARAMETER_REGISTERS, REGISTER_A
from ...context.stack.helpers import (
    enter_new_stack,
    exit_stack,
    get_datastack_location,
    silently_allocate_on_datastack,
)
from ...debugstatement import debug_statement
from ...types.primitives.void import VoidType
from ..expression import Expression


class FunctionCall(Expression):
    def __init__(self, name: str, parameters: list[Expression]) -> None:
        super().__init__()
        self._name = name
        self._parameters = parameters

    def __repr__(self):
        return f"{self._name}(" + ", ".join(repr(p) for p in self._parameters) + ")"

    def get_type(self, context):
        function_def = context.function_registry.get_function_definition(self._name)
        if function_def:
            return function_def.return_type
        else:
            return VoidType()

    def evaluate(self, context: EvaluationContext) -> str:
        function_def = context.function_registry.get_function_definition(self._name)
        if not function_def:
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Function '{self._name}' is not defined.",
            )
            return ""

        if len(function_def.parameters) != len(self._parameters):
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Function '{self._name}' expects {len(function_def.parameters)} parameters, "
                f"but {len(self._parameters)} were given.",
            )
            return ""

        for i in range(min(len(self._parameters), len(function_def.parameters))):
            expected_type = function_def.parameters[i].type
            actual_type = self._parameters[i].get_type(context)
            if actual_type != expected_type:
                context.register_issue(
                    IssueLevel.ERROR,
                    self,
                    f"Parameter {i + 1} of function '{self._name}' expects type "
                    f"'{expected_type}', but got type '{actual_type}'.",
                )
                return ""

        if len(function_def.parameters) > len(FUNCTION_PARAMETER_REGISTERS):
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Function '{self._name}' has too many parameters "
                f"(max {len(FUNCTION_PARAMETER_REGISTERS)}), "
                f"currently only functions with up to {len(FUNCTION_PARAMETER_REGISTERS)} "
                "parameters are supported.",
            )
            return ""

        function_label = context.function_registry.get_function_label(self._name)

        res = ""
        res += debug_statement("Function Call: " + self._name)

        # Evaluate parameters and store them in registers
        for i in range(len(self._parameters)):
            if self._parameters[i].get_type(context).get_size() != 1:
                context.register_issue(
                    IssueLevel.ERROR,
                    self,
                    "Currently only parameters of size 1 are supported.",
                )
                return ""

            res += PUSH(FUNCTION_PARAMETER_REGISTERS[i])
            with increment_stack(context):
                res += self._parameters[i].evaluate(context)
            res += STORE

        # Set up new stack frame
        res += enter_new_stack(context)
        with reset_stack(context):
            # Allocate space for function variables
            total_var_size = sum(v.type.get_size() for v in function_def.variables)
            res += silently_allocate_on_datastack(context, total_var_size)

            # Copy parameters into variables in new stack frame
            function_def.assert_correct_variable_configuration()
            for i, param in enumerate(function_def.parameters):
                matching_var = function_def.get_variable(param.name)
                if not matching_var:
                    continue  # This should not happen due to earlier checks
                res += get_datastack_location(context, matching_var.offset)
                res += PUSH(FUNCTION_PARAMETER_REGISTERS[i]) + FETCH + STORE

            # Call the function
            res += CALL(function_label)

            res += debug_statement("Returned from function: " + self._name)

            # Safely store return value
            res += PUSH(REGISTER_A) + SWAP + STORE

        # Dispose of stack frame
        res += exit_stack(context)

        # Move return value back to stack
        res += PUSH(REGISTER_A) + FETCH

        return res



from ...whitespace.snippets import STORE
from ..context import EvaluationContext, IssueLevel
from ..context.evaluation_context import increment_stack
from ..expressions.expression import Expression
from ..types.complex.memory import MemoryType
from ..utils.error_message import type_error
from .statement import Statement


class AssignmentStatement(Statement):
    def __init__(self, target: Expression, value: Expression) -> None:
        super().__init__()
        self._target = target
        self._value = value

    def __repr__(self):
        return f"{self._target} = {self._value}"

    def evaluate(self, context: EvaluationContext) -> str:
        target_type = self._target.get_type(context)
        if not isinstance(target_type, MemoryType):
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Invalid assignment target: cannot assign to expression of type '{target_type}'.",
            )
            return ""

        inner_target_type = target_type.of
        value_type = self._value.get_type(context)
        if inner_target_type != value_type:
            context.register_issue(
                IssueLevel.ERROR,
                self,
                type_error(inner_target_type, value_type, self),
            )
            return ""

        res = self._target.evaluate(context)
        with increment_stack(context):
            res += self._value.evaluate(context)
        res += STORE

        return res



from ...context import EvaluationContext, IssueLevel
from ...context.stack.helpers import get_datastack_location
from ...definitions.function import VariableOffset
from ...types.complex.memory import MemoryType
from ...types.primitives.void import VoidType
from ...types.type import Type
from ..expression import Expression


class VariableAccess(Expression):
    def __init__(self, name: str):
        super().__init__()
        self._name = name

    def __repr__(self):
        return self._name

    def _get_variable_definition(self, context: EvaluationContext) -> VariableOffset | None:
        if not context.function_registry.current_function:
            return None
        return context.function_registry.current_function.get_variable(self._name)

    def get_type(self, context: EvaluationContext) -> Type:
        definition = self._get_variable_definition(context)
        if definition is None:
            return VoidType()
        return MemoryType(definition.definition.type)

    def evaluate(self, context: EvaluationContext) -> str:
        definition = self._get_variable_definition(context)
        if not definition:
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Variable '{self._name}' is not defined.",
            )
            return ""

        res = get_datastack_location(context, definition.offset)
        return res



REGISTER_A = 0
REGISTER_B = 1
REGISTER_C = 2
REGISTER_D = 3

_max_parameter_count = 15
_function_parameter_start = REGISTER_D + 1
FUNCTION_PARAMETER_REGISTERS = list(
    range(_function_parameter_start, _function_parameter_start + _max_parameter_count)
)

FIRST_PROGRAM_STACK_HEAP_ADDRESS = max(FUNCTION_PARAMETER_REGISTERS) + 1
PROGRAM_STACK_LIMIT = FIRST_PROGRAM_STACK_HEAP_ADDRESS + 2048 * 128

FREE_HEAP_ALLOCATION_COUNTER = PROGRAM_STACK_LIMIT + 1
FREE_HEAP_START = FREE_HEAP_ALLOCATION_COUNTER + 1



from ....whitespace.printstr import print_str
from ....whitespace.snippets import ADD, DUPLICATE, PRINT_NUMBER, PUSH
from ...context import EvaluationContext, IssueLevel
from ...context.evaluation_context import increment_stack
from ...debugstatement import debug_instructions
from ...types.complex.array import ArrayType
from ...types.complex.memory import MemoryType
from ...types.complex.string import StringType
from ...types.primitives.int import IntegerType
from ...types.primitives.void import VoidType
from ...types.type import Type
from ...utils.error_message import type_error
from ..expression import Expression


class ArrayIndex(Expression):
    def __init__(self, array: Expression, index: Expression) -> None:
        super().__init__()
        self._array = array
        self._index = index

    def __repr__(self) -> str:
        return f"{self._array}[{self._index}]"

    def get_type(self, context: EvaluationContext) -> Type:
        arr_type = self._array.get_type(context)

        if isinstance(arr_type, ArrayType):
            return MemoryType(arr_type.of)
        elif isinstance(arr_type, StringType):
            return MemoryType(IntegerType())

        return VoidType()

    def evaluate(self, context: EvaluationContext) -> str:
        arr_type = self._array.get_type(context)
        if not isinstance(arr_type, (ArrayType, StringType)):
            context.register_issue(
                IssueLevel.ERROR,
                self._array,
                f"Type '{arr_type}' is not indexable.",
            )
            return ""

        index_type = self._index.get_type(context)
        if index_type != IntegerType():
            context.register_issue(
                IssueLevel.ERROR,
                self,
                type_error(IntegerType(), index_type, self._index),
            )
            return ""

        res = ""
        res += self._array.evaluate(context)
        with increment_stack(context):
            res += self._index.evaluate(context)
        res += PUSH(1) + ADD + ADD

        res += debug_instructions(
            print_str("Accessing array item at ") + DUPLICATE + PRINT_NUMBER + print_str("\n")
        )

        return res



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from typing import Callable

from ..node import AstNode
from .function_registry import FunctionRegistry
from .issue_level import IssueLevel
from .label_registry import LabelRegistry


@dataclass
class Issue:
    level: IssueLevel
    node: AstNode
    message: str


@contextmanager
def increment_stack(context: EvaluationContext, by: int = 1):
    try:
        context._stack_offset += by
        yield
    finally:
        context._stack_offset -= by


@contextmanager
def reset_stack(context: EvaluationContext):
    try:
        original_offset = context._stack_offset
        context._stack_offset = 0
        yield
    finally:
        if context._stack_offset != 0:
            raise RuntimeError("Stack offset must be zero when exiting reset_stack context")
        context._stack_offset = original_offset


class EvaluationContext:
    def __init__(self) -> None:
        self.label_registry = LabelRegistry()
        self.function_registry = FunctionRegistry()

        self._issues: list[Issue] = []
        self._stack_offset = 0

    def register_issue(self, level: IssueLevel, node: AstNode, message: str) -> None:
        self._issues.append(Issue(level, node, message))

    def with_incremented_stack(self, fn: Callable[[EvaluationContext], str]) -> str:
        with increment_stack(self):
            return fn(self)

    @property
    def stack_offset(self) -> int:
        return self._stack_offset

    def print_issues(self, strict: bool = False) -> None:
        has_warnings = False
        has_errors = False

        for issue in self._issues:
            print(f"[{issue.level.name}] {issue.message} (at {issue.node})")

            if issue.level == IssueLevel.ERROR:
                has_errors = True
            elif issue.level == IssueLevel.WARNING:
                has_warnings = True

        if has_errors:
            raise RuntimeError("Compilation failed due to errors.")
        if strict and has_warnings:
            raise RuntimeError("Compilation failed due to warnings in strict mode.")



from ....whitespace.snippets import JUMP, JUMP_IF_ZERO, LABEL
from ...context import EvaluationContext
from ...types.primitives.bool import BooleanType
from ..literals.bool import FalseLiteral, TrueLiteral
from .operator import BinaryOperator, UnaryOperator


class NotOperator(UnaryOperator):
    def __repr__(self):
        return f"!{self.inner_expression}"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_inner_as_type(context, BooleanType())

        false_label = context.label_registry.new_label()
        end_label = context.label_registry.new_label()

        return (
            self.inner_expression.evaluate(context)
            + JUMP_IF_ZERO(false_label)
            + FalseLiteral().evaluate(context)
            + JUMP(end_label)
            + LABEL(false_label)
            + TrueLiteral().evaluate(context)
            + LABEL(end_label)
        )


class AndOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} && {self.right_side})"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, BooleanType())

        false_label = context.label_registry.new_label()
        end_label = context.label_registry.new_label()

        return (
            self.left_side.evaluate(context)
            + JUMP_IF_ZERO(false_label)
            + self.right_side.evaluate(context)
            + JUMP_IF_ZERO(false_label)
            + TrueLiteral().evaluate(context)
            + JUMP(end_label)
            + LABEL(false_label)
            + FalseLiteral().evaluate(context)
            + LABEL(end_label)
        )


class OrOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} || {self.right_side})"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, BooleanType())

        check_right_side_label = context.label_registry.new_label()
        false_label = context.label_registry.new_label()
        end_label = context.label_registry.new_label()

        return (
            self.left_side.evaluate(context)
            + JUMP_IF_ZERO(check_right_side_label)
            + TrueLiteral().evaluate(context)
            + JUMP(end_label)
            + LABEL(check_right_side_label)
            + self.right_side.evaluate(context)
            + JUMP_IF_ZERO(false_label)
            + TrueLiteral().evaluate(context)
            + JUMP(end_label)
            + LABEL(false_label)
            + FalseLiteral().evaluate(context)
            + LABEL(end_label)
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from ....whitespace.snippets import ADD, DUPLICATE, FETCH, PUSH, STORE, SWAP
from ...debugstatement import debug_statement
from .consts import FREE_HEAP_ALLOCATION_COUNTER, FREE_HEAP_START

if TYPE_CHECKING:
    from ..evaluation_context import EvaluationContext


def initialize_heap(context: EvaluationContext) -> str:
    """
    Initialize the heap by setting up the heap allocation counter.
    This function should be called once at the start of the program execution.
    """
    _ = context  # Unused
    return (
        debug_statement("Initializing heap")
        + PUSH(FREE_HEAP_ALLOCATION_COUNTER)
        + PUSH(FREE_HEAP_START)
        + STORE
    )


def place_const_data_on_heap(context: EvaluationContext, data: list[int]) -> str:
    """
    Place constant data on the heap and return the starting address.
    This can be used for string literals or other constant data.
    The context stack offset is NOT incremented, that is the responsibility of the caller.
    """
    _ = context  # Unused

    # get the current free heap location address
    res = (
        debug_statement(f"Placing {len(data)} long const data on heap")
        + PUSH(FREE_HEAP_ALLOCATION_COUNTER)
        + FETCH
        + DUPLICATE
    )

    data = [len(data)] + data  # prepend length

    for i in range(len(data)):
        res += DUPLICATE
        res += PUSH(data[i]) + STORE
        res += PUSH(1) + ADD
    res += PUSH(FREE_HEAP_ALLOCATION_COUNTER) + SWAP + STORE
    return res


def allocate_data_on_heap(context: EvaluationContext, size: int) -> str:
    """
    Allocate space on the heap for a given size and return the starting address.
    The first word at that address will contain the size of the allocated block.
    The context stack offset is NOT incremented, that is the responsibility of the caller.
    """
    _ = context  # Unused

    # get the current free heap location address
    res = (
        debug_statement(f"Allocating {size} long space on heap")
        + PUSH(FREE_HEAP_ALLOCATION_COUNTER)
        + FETCH
    )

    # on the free location, first store the size
    res += DUPLICATE + PUSH(size) + STORE

    # increment the free heap location by size + 1
    res += DUPLICATE + PUSH(size + 1) + ADD
    res += PUSH(FREE_HEAP_ALLOCATION_COUNTER) + SWAP + STORE

    return res



from .converters import to_ws_number
from .tokens import LF, SPACE, TAB

PUSH = lambda n: SPACE + SPACE + to_ws_number(n)  # noqa: E731
POP = SPACE + LF + LF
SWAP = SPACE + LF + TAB
DUPLICATE = SPACE + LF + SPACE
COPY = lambda n: SPACE + TAB + SPACE + to_ws_number(n)  # noqa: E731
SLIDE = lambda n: SPACE + TAB + LF + to_ws_number(n)  # noqa: E731

ADD = TAB + SPACE + SPACE + SPACE
SUBTRACT = TAB + SPACE + SPACE + TAB
MULTIPLY = TAB + SPACE + SPACE + LF
DIVIDE = TAB + SPACE + TAB + SPACE
MODULO = TAB + SPACE + TAB + TAB

FETCH = TAB + TAB + TAB
STORE = TAB + TAB + SPACE

LABEL = lambda n: LF + SPACE + SPACE + to_ws_number(n)  # noqa: E731
CALL = lambda n: LF + SPACE + TAB + to_ws_number(n)  # noqa: E731
JUMP = lambda n: LF + SPACE + LF + to_ws_number(n)  # noqa: E731
JUMP_IF_ZERO = lambda n: LF + TAB + SPACE + to_ws_number(n)  # noqa: E731
JUMP_IF_NEG = lambda n: LF + TAB + TAB + to_ws_number(n)  # noqa: E731
RETURN = LF + TAB + LF
END = LF + LF + LF

PRINT_CHAR = TAB + LF + SPACE + SPACE
PRINT_NUMBER = TAB + LF + SPACE + TAB
READ_CHAR = TAB + LF + TAB + SPACE
READ_NUMBER = TAB + LF + TAB + TAB



from ....whitespace.snippets import ADD, DIVIDE, MODULO, MULTIPLY, SUBTRACT
from ...context import EvaluationContext
from ...types.primitives.int import IntegerType
from .operator import BinaryOperator


class AddOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} + {self.right_side})"

    def get_type(self, context):
        return IntegerType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())
        return self._evaluate_both_sides(context) + ADD


class SubtractOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} - {self.right_side})"

    def get_type(self, context):
        return IntegerType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())
        return self._evaluate_both_sides(context) + SUBTRACT


class MultiplyOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} * {self.right_side})"

    def get_type(self, context):
        return IntegerType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())
        return self._evaluate_both_sides(context) + MULTIPLY


class DivideOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} / {self.right_side})"

    def get_type(self, context):
        return IntegerType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())
        return self._evaluate_both_sides(context) + DIVIDE


class ModuloOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} % {self.right_side})"

    def get_type(self, context):
        return IntegerType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())
        return self._evaluate_both_sides(context) + MODULO



from ..whitespace.printstr import print_str

ENABLE_DEBUG = False


def debug_statement(text: str) -> str:
    if not ENABLE_DEBUG:
        return ""
    return (
        code_comment(f"DEBUG: {text}")
        + print_str(f"DEBUG: {text}\n")
        + code_comment("// end of debug statement print")
    )


def debug_instructions(instructions: str) -> str:
    if not ENABLE_DEBUG:
        return ""
    return instructions


def code_comment(text: str) -> str:
    if "\n" in text:
        raise ValueError("Code comments cannot contain newlines")
    if "\t" in text:
        raise ValueError("Code comments cannot contain tabs")
    return text.replace(" ", "·")



from ..context import EvaluationContext
from ..utils.indent import indent
from .statement import Statement


class CompoundStatement(Statement):
    """A compound statement is a sequence of statements executed in order."""

    def __init__(self, statements: list[Statement]) -> None:
        super().__init__()
        self.statements = statements

    def __repr__(self):
        res = "{\n"
        for stmt in self.statements:
            res += indent(repr(stmt)) + "\n"
        res += "}"
        return res

    def evaluate(self, context: EvaluationContext):
        code = ""
        for statement in self.statements:
            code += statement.evaluate(context)
        return code



from ....whitespace.snippets import JUMP, JUMP_IF_NEG, JUMP_IF_ZERO, LABEL, SUBTRACT, SWAP
from ...context import EvaluationContext
from ...types.primitives.bool import BooleanType
from ...types.primitives.int import IntegerType
from ..literals.bool import FalseLiteral, TrueLiteral
from .operator import BinaryOperator


class EqualOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} == {self.right_side})"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_same_type(context)

        label_equals = context.label_registry.new_label()
        label_end = context.label_registry.new_label()

        return (
            self._evaluate_both_sides(context)
            + SUBTRACT
            + JUMP_IF_ZERO(label_equals)
            + FalseLiteral().evaluate(context)
            + JUMP(label_end)
            + LABEL(label_equals)
            + TrueLiteral().evaluate(context)
            + LABEL(label_end)
        )


class NotEqualOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} != {self.right_side})"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_same_type(context)

        label_equals = context.label_registry.new_label()
        label_end = context.label_registry.new_label()

        return (
            self._evaluate_both_sides(context)
            + SUBTRACT
            + JUMP_IF_ZERO(label_equals)
            + TrueLiteral().evaluate(context)
            + JUMP(label_end)
            + LABEL(label_equals)
            + FalseLiteral().evaluate(context)
            + LABEL(label_end)
        )


class GreaterThanOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} > {self.right_side})"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())

        label_greater = context.label_registry.new_label()
        label_end = context.label_registry.new_label()

        return (
            self._evaluate_both_sides(context)
            + SWAP
            + SUBTRACT
            + JUMP_IF_NEG(label_greater)
            + FalseLiteral().evaluate(context)
            + JUMP(label_end)
            + LABEL(label_greater)
            + TrueLiteral().evaluate(context)
            + LABEL(label_end)
        )


class GreaterOrEqualThanOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} >= {self.right_side})"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())

        label_less = context.label_registry.new_label()
        label_end = context.label_registry.new_label()

        return (
            self._evaluate_both_sides(context)
            + SUBTRACT
            + JUMP_IF_NEG(label_less)
            + TrueLiteral().evaluate(context)
            + JUMP(label_end)
            + LABEL(label_less)
            + FalseLiteral().evaluate(context)
            + LABEL(label_end)
        )


class LessThanOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} < {self.right_side})"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())

        label_less = context.label_registry.new_label()
        label_end = context.label_registry.new_label()

        return (
            self._evaluate_both_sides(context)
            + SUBTRACT
            + JUMP_IF_NEG(label_less)
            + FalseLiteral().evaluate(context)
            + JUMP(label_end)
            + LABEL(label_less)
            + TrueLiteral().evaluate(context)
            + LABEL(label_end)
        )


class LessOrEqualThanOperator(BinaryOperator):
    def __repr__(self):
        return f"({self.left_side} <= {self.right_side})"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext):
        self._enforce_both_sides_as_type(context, IntegerType())

        label_greater = context.label_registry.new_label()
        label_end = context.label_registry.new_label()

        return (
            self._evaluate_both_sides(context)
            + SWAP
            + SUBTRACT
            + JUMP_IF_NEG(label_greater)
            + TrueLiteral().evaluate(context)
            + JUMP(label_end)
            + LABEL(label_greater)
            + FalseLiteral().evaluate(context)
            + LABEL(label_end)
        )



from ...whitespace.snippets import POP
from ..context import EvaluationContext
from ..expressions.expression import Expression
from .statement import Statement


class VoidStatement(Statement):
    """A void statement evaluates an expression and discards its result."""

    def __init__(self, expression: Expression) -> None:
        super().__init__()
        self.expression = expression

    def evaluate(self, context: EvaluationContext):
        return self.expression.evaluate(context) + POP



from contextlib import contextmanager

from ..definitions.function import FunctionSignature


class FunctionRegistry:
    def __init__(self) -> None:
        self._functions: list[FunctionSignature] = []
        self._labels: dict[str, int] = {}
        self._current_function: FunctionSignature | None = None

    def register_function(self, function: FunctionSignature, label_id: int) -> None:
        if any(fn.name == function.name for fn in self._functions):
            raise RuntimeError(f"Function '{function.name}' is already registered.")

        self._functions.append(function)
        self._labels[function.name] = label_id

    def get_function_definition(self, function_name: str) -> FunctionSignature | None:
        for function in self._functions:
            if function.name == function_name:
                return function
        return None

    def get_function_label(self, function_name: str) -> int:
        """
        Get the label ID for the given function name.
        Should not be called unless the function is known to exist by using
        `get_function_definition` first.
        """
        return self._labels[function_name]

    @contextmanager
    def function_context(self, function: FunctionSignature):
        if self._current_function is not None:
            raise RuntimeError("Nested function definitions are not supported.")
        self._current_function = function
        try:
            yield
        finally:
            self._current_function = None

    @property
    def current_function(self) -> FunctionSignature | None:
        return self._current_function



from ....whitespace.printstr import print_str
from ....whitespace.snippets import ADD, DUPLICATE, POP, PRINT_NUMBER, PUSH, STORE
from ...context import EvaluationContext, IssueLevel
from ...context.evaluation_context import increment_stack
from ...context.heap.helpers import allocate_data_on_heap
from ...debugstatement import debug_instructions, debug_statement
from ...types.complex.array import ArrayType
from ...types.primitives.void import VoidType
from ...types.type import Type
from ...utils.error_message import type_error
from ..expression import Expression


class ArrayLiteral(Expression):
    def __init__(self, items: list[Expression]) -> None:
        super().__init__()
        self._items = items

    def __repr__(self):
        return "[" + ", ".join(repr(item) for item in self._items) + "]"

    def get_type(self, context: EvaluationContext) -> Type:
        inner_type = self._items[0].get_type(context) if self._items else VoidType()
        return ArrayType(inner_type)

    def evaluate(self, context: EvaluationContext):
        first_item_type = self._items[0].get_type(context) if self._items else VoidType()
        for item in self._items:
            item_type = item.get_type(context)
            if item_type != first_item_type:
                context.register_issue(
                    IssueLevel.ERROR,
                    self,
                    type_error(first_item_type, item_type, item),
                )

        # get the starting location of the heap
        res = allocate_data_on_heap(context, len(self._items))

        res += debug_statement(f"Allocating array with {len(self._items)} items")

        res += DUPLICATE
        for i, item in enumerate(self._items):
            res += PUSH(1) + ADD + DUPLICATE

            res += debug_instructions(
                print_str("[") + DUPLICATE + PRINT_NUMBER + print_str("] = ")
            )

            with increment_stack(context, 3):
                res += item.evaluate(context)

            res += debug_instructions(DUPLICATE + PRINT_NUMBER + print_str("\n"))

            res += STORE
        res += POP

        return res



from abc import ABC, abstractmethod

from ..context import EvaluationContext
from ..node import AstNode


class Statement(AstNode, ABC):
    @abstractmethod
    def evaluate(self, context: EvaluationContext) -> str:
        pass



from abc import ABC

from ...whitespace.snippets import JUMP, JUMP_IF_ZERO, LABEL
from ..context import EvaluationContext, IssueLevel
from ..expressions.expression import Expression
from ..types.primitives.bool import BooleanType
from ..utils.error_message import type_error
from .statement import Statement


class LoopStatement(Statement, ABC):
    def __init__(self):
        pass


class WhileLoopStatement(LoopStatement):
    def __init__(self, condition: Expression, body: Statement):
        super().__init__()
        self.condition = condition
        self.body = body

    def __repr__(self):
        return f"while {self.condition}: {self.body}"

    def evaluate(self, context: EvaluationContext) -> str:
        condition_type = self.condition.get_type(context)
        if condition_type != BooleanType():
            context.register_issue(
                IssueLevel.ERROR,
                self,
                type_error(BooleanType(), condition_type, self),
            )

        loop_start_label = context.label_registry.new_label()
        loop_end_label = context.label_registry.new_label()

        return (
            LABEL(loop_start_label)
            + self.condition.evaluate(context)
            + JUMP_IF_ZERO(loop_end_label)
            + self.body.evaluate(context)
            + JUMP(loop_start_label)
            + LABEL(loop_end_label)
        )


class DoWhileStatement(LoopStatement):
    def __init__(self, condition: Expression, body: Statement):
        super().__init__()
        self.condition = condition
        self.body = body

    def __repr__(self):
        return f"do {self.body} while {self.condition}"

    def evaluate(self, context: EvaluationContext) -> str:
        condition_type = self.condition.get_type(context)
        if condition_type != BooleanType():
            context.register_issue(
                IssueLevel.ERROR,
                self,
                type_error(BooleanType(), condition_type, self),
            )

        loop_start_label = context.label_registry.new_label()
        loop_end_label = context.label_registry.new_label()

        return (
            LABEL(loop_start_label)
            + self.body.evaluate(context)
            + self.condition.evaluate(context)
            + JUMP_IF_ZERO(loop_end_label)
            + JUMP(loop_start_label)
            + LABEL(loop_end_label)
        )


class ForLoopStatement(LoopStatement):
    def __init__(
        self,
        initializer: Statement,
        condition: Expression,
        increment: Statement,
        body: Statement,
    ):
        super().__init__()
        self.initializer = initializer
        self.condition = condition
        self.increment = increment
        self.body = body

    def __repr__(self):
        return f"for ({self.initializer}; {self.condition}; {self.increment}) {self.body}"

    def evaluate(self, context: EvaluationContext) -> str:
        condition_type = self.condition.get_type(context)
        if condition_type != BooleanType():
            context.register_issue(
                IssueLevel.ERROR,
                self,
                type_error(BooleanType(), condition_type, self),
            )

        loop_start_label = context.label_registry.new_label()
        loop_end_label = context.label_registry.new_label()

        return (
            self.initializer.evaluate(context)
            + LABEL(loop_start_label)
            + self.condition.evaluate(context)
            + JUMP_IF_ZERO(loop_end_label)
            + self.body.evaluate(context)
            + self.increment.evaluate(context)
            + JUMP(loop_start_label)
            + LABEL(loop_end_label)
        )



from .tokens import LF, SPACE, TAB


def to_ws_number(value: int) -> str:
    ret = ""
    if value >= 0:
        ret += SPACE
    else:
        ret += TAB
        value = -value

    binary = bin(value)[2:]
    for bit in binary:
        if bit == "0":
            ret += SPACE
        else:
            ret += TAB

    ret += LF

    return ret



from __future__ import annotations

from typing import TYPE_CHECKING

from ....whitespace.snippets import ADD, COPY, DUPLICATE, FETCH, POP, PUSH, STORE
from ...debugstatement import debug_statement
from ..heap.consts import FIRST_PROGRAM_STACK_HEAP_ADDRESS

if TYPE_CHECKING:
    from ..evaluation_context import EvaluationContext


# A function stack is a stack frame used during the evaluation of functions.
# When entering a function, a new stack frame is created to hold local variables and state.
# When exiting a function, the stack frame is removed, and control returns to the previous stack.
# The data of the stack frames is stored on the runtime heap.
# The stack data store location is pushed to the whitespace stack.


def _get_datastack_location(context: EvaluationContext) -> str:
    """Get the heap location of the current data stack."""
    return COPY(context.stack_offset)


def enter_first_stack(context: EvaluationContext) -> str:
    """Enter the first stack of the execution context."""
    _ = context  # Unused
    return (
        debug_statement("Entering first stack")
        + PUSH(FIRST_PROGRAM_STACK_HEAP_ADDRESS)
        + DUPLICATE
        + PUSH(1)
        + STORE
    )


def enter_new_stack(context: EvaluationContext) -> str:
    """Enter a new stack in the execution context. Assumes there is already at least one stack."""
    # Get the location of the current stack, get the length of the current stack,
    # add them together to the the start of the new stack.
    # Push 1 to the new stack to indicate its length is starting at 1.
    res = debug_statement(f"Entering new stack with stack offset {context.stack_offset}")
    res += _get_datastack_location(context)
    res += DUPLICATE + FETCH + ADD
    res += DUPLICATE + PUSH(1) + STORE
    return res


def exit_stack(context: EvaluationContext) -> str:
    """Exit the current stack in the execution context."""
    _ = context  # Unused
    return debug_statement(f"Exiting stack with stack offset {context.stack_offset}") + POP


def silently_allocate_on_datastack(context: EvaluationContext, length: int) -> str:
    """
    Allocate data on the current data stack.
    The starting location of the data is NOT pushed onto the whitespace stack.
    """
    res = _get_datastack_location(context)
    res += DUPLICATE + FETCH + PUSH(length) + ADD + STORE
    return res


def get_datastack_location(context: EvaluationContext, offset: int) -> str:
    """
    Get the heap location of the current data stack plus the given offset.
    **Skips the first element of the data stack, which is the length of the stack.**
    The resulting location is pushed onto the whitespace stack.
    The context stack offset is NOT incremented, that is the responsibility of the caller.
    """
    offset += 1  # Skip the length element
    res = _get_datastack_location(context)
    if offset != 0:
        res += PUSH(offset) + ADD
    return res



from ....whitespace.snippets import PUSH
from ...context import EvaluationContext
from ...types.primitives.int import IntegerType
from ..expression import Expression


class IntLiteral(Expression):

    def __init__(self, value: int) -> None:
        super().__init__()
        self.value = value

    def __repr__(self):
        return str(self.value)

    def get_type(self, context):
        return IntegerType()

    def evaluate(self, context: EvaluationContext) -> str:
        return PUSH(self.value)



from ..type import Type


class StringType(Type):
    def __repr__(self):
        return "String"

    def __eq__(self, value):
        return isinstance(value, StringType)

    def get_size(self):
        return 1



from ....whitespace.snippets import PUSH
from ...context import EvaluationContext
from ...types.primitives.bool import BooleanType
from ..expression import Expression


class TrueLiteral(Expression):

    def __repr__(self):
        return "TRUE"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext) -> str:
        return PUSH(1)


class FalseLiteral(Expression):

    def __repr__(self):
        return "FALSE"

    def get_type(self, context):
        return BooleanType()

    def evaluate(self, context: EvaluationContext) -> str:
        return PUSH(0)



from enum import Enum

from ...whitespace.snippets import (
    ADD,
    COPY,
    DUPLICATE,
    FETCH,
    JUMP,
    JUMP_IF_ZERO,
    LABEL,
    POP,
    PRINT_CHAR,
    PRINT_NUMBER,
    PUSH,
    SUBTRACT,
    SWAP,
)
from ..context import EvaluationContext, IssueLevel
from ..expressions.expression import Expression
from ..types.complex.array import ArrayType
from ..types.complex.memory import MemoryType
from ..types.complex.string import StringType
from ..types.primitives.bool import BooleanType
from ..types.primitives.int import IntegerType
from .statement import Statement


class PrintMode(Enum):
    Number = 1
    Character = 2


class PrintStatement(Statement):

    def __init__(self, expression: Expression, print_mode: PrintMode = PrintMode.Number) -> None:
        super().__init__()
        self._expression = expression
        self._print_mode = print_mode

    def __repr__(self) -> str:
        return f"Print {self._expression} as {self._print_mode}"

    def evaluate(self, context: EvaluationContext) -> str:
        expr_type = self._expression.get_type(context)

        res = ""

        if expr_type == IntegerType() or expr_type == BooleanType():
            res += self._expression.evaluate(context)

            if self._print_mode == PrintMode.Character:
                res += PRINT_CHAR
            else:
                res += PRINT_NUMBER

        elif expr_type == StringType() or isinstance(expr_type, ArrayType):
            res += self._expression.evaluate(context)
            # In array/string types, the evaluated data points to the heap address where
            # the actual data is stored. The first location on the heap is the length of the data,
            # followed by the actual data items.

            res += DUPLICATE + DUPLICATE + FETCH + PUSH(1) + ADD + ADD
            # Stack: [start_address, end_address_exclusive]

            print_loop_start_label = context.label_registry.new_label()
            print_loop_end_label = context.label_registry.new_label()

            res += LABEL(print_loop_start_label)

            # increment the start address
            res += SWAP + PUSH(1) + ADD + SWAP

            # check if we've reached the end
            res += COPY(1) + COPY(1) + SUBTRACT + JUMP_IF_ZERO(print_loop_end_label)

            # fetch and print the current item
            res += COPY(1) + FETCH

            if self._print_mode == PrintMode.Character:
                res += PRINT_CHAR
            else:
                res += PRINT_NUMBER

            res += JUMP(print_loop_start_label)
            res += LABEL(print_loop_end_label)
            res += POP + POP

        elif isinstance(expr_type, MemoryType):
            res += self._expression.evaluate(context)
            res += PRINT_NUMBER

        else:
            context.register_issue(
                IssueLevel.ERROR,
                self,
                f"Cannot print expression of type {expr_type}",
            )

        return res



from abc import ABC

from ...context import EvaluationContext, IssueLevel
from ...types.type import Type
from ...utils.error_message import type_error
from ..expression import Expression


class Operator(Expression, ABC):
    pass


class UnaryOperator(Operator, ABC):
    def __init__(self, inner_expr: Expression):
        super().__init__()
        self.inner_expression = inner_expr

    def _enforce_inner_as_type(self, context: EvaluationContext, type: Type) -> None:
        inner_type = self.inner_expression.get_type(context)
        if inner_type != type:
            context.register_issue(
                IssueLevel.ERROR,
                self.inner_expression,
                type_error(type, inner_type, self),
            )


class BinaryOperator(Operator, ABC):
    def __init__(self, left_side: Expression, right_side: Expression):
        super().__init__()
        self.left_side = left_side
        self.right_side = right_side

    def _evaluate_both_sides(self, context: EvaluationContext) -> str:
        return self.left_side.evaluate(context) + context.with_incremented_stack(
            lambda c: self.right_side.evaluate(c)
        )

    def _enforce_both_sides_as_type(self, context: EvaluationContext, type: Type) -> None:
        left_side_type = self.left_side.get_type(context)
        if self.left_side.get_type(context) != type:
            context.register_issue(
                IssueLevel.ERROR,
                self.left_side,
                type_error(type, left_side_type, self.left_side),
            )

        right_side_type = self.right_side.get_type(context)
        if self.right_side.get_type(context) != type:
            context.register_issue(
                IssueLevel.ERROR,
                self.right_side,
                type_error(type, right_side_type, self.right_side),
            )

    def _enforce_both_sides_as_same_type(self, context: EvaluationContext) -> None:
        left_side_type = self.left_side.get_type(context)
        right_side_type = self.right_side.get_type(context)
        if left_side_type != right_side_type:
            context.register_issue(
                IssueLevel.ERROR,
                self,
                "Type mismatch between left and right side of operation: "
                f"{left_side_type} vs {right_side_type}",
            )



from ....whitespace.snippets import FETCH
from ...context import EvaluationContext, IssueLevel
from ...types.complex.memory import MemoryType
from ...types.primitives.void import VoidType
from ...types.type import Type
from ..expression import Expression


class MemoryFetch(Expression):
    def __init__(self, inner_expr) -> None:
        super().__init__()
        self._inner_expr = inner_expr

    def __repr__(self) -> str:
        return f"resolve({self._inner_expr})"

    def get_type(self, context: EvaluationContext) -> Type:
        inner_type = self._inner_expr.get_type(context)
        if isinstance(inner_type, MemoryType):
            return inner_type.of
        return VoidType()

    def evaluate(self, context: EvaluationContext) -> str:
        arr_type = self._inner_expr.get_type(context)
        if not isinstance(arr_type, MemoryType):
            context.register_issue(
                IssueLevel.ERROR,
                self._inner_expr,
                f"Type '{arr_type}' is not dereferenceable.",
            )
            return ""

        res = ""
        res += self._inner_expr.evaluate(context)
        res += FETCH

        return res



from ..type import Type


class MemoryType(Type):
    def __init__(self, of: Type):
        super().__init__()
        self.of = of

    def __repr__(self):
        return f"Memory<{self.of}>"

    def __eq__(self, value):
        return isinstance(value, MemoryType) and self.of == value.of

    def get_size(self):
        return 1



from __future__ import annotations

from ...whitespace.snippets import JUMP, JUMP_IF_ZERO, LABEL
from ..context import EvaluationContext, IssueLevel
from ..expressions.expression import Expression
from ..types.primitives.bool import BooleanType
from ..utils.error_message import type_error
from ..utils.indent import indent
from .statement import Statement


class ConditionStatement(Statement):
    def __init__(
        self, condition: Expression, if_true: Statement, if_false: Statement | None = None
    ):
        super().__init__()
        self.condition = condition
        self.if_true = if_true
        self.if_false = if_false

    def __repr__(self) -> str:
        res = f"if {self.condition} then\n"
        res += indent(repr(self.if_true))
        if self.if_false:
            res += f"\nelse\n{indent(repr(self.if_false))}"
        return res

    def evaluate(self, context: EvaluationContext):
        condition_type = self.condition.get_type(context)
        if condition_type != BooleanType():
            context.register_issue(
                IssueLevel.ERROR,
                self,
                type_error(BooleanType(), condition_type, self),
            )

        res = self.condition.evaluate(context)
        if self.if_false:
            if_false_label = context.label_registry.new_label()
            end_label = context.label_registry.new_label()

            res += (
                JUMP_IF_ZERO(if_false_label)
                + self.if_true.evaluate(context)
                + JUMP(end_label)
                + LABEL(if_false_label)
                + self.if_false.evaluate(context)
                + LABEL(end_label)
            )
        else:
            end_label = context.label_registry.new_label()

            res += JUMP_IF_ZERO(end_label) + self.if_true.evaluate(context) + LABEL(end_label)

        return res
